# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("snmpapi.dll")
prototypes = \
    {
        #
        'SnmpUtilOidCpy': SimTypeFunction([SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pOidDst", "pOidSrc"]),
        #
        'SnmpUtilOidAppend': SimTypeFunction([SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pOidDst", "pOidSrc"]),
        #
        'SnmpUtilOidNCmp': SimTypeFunction([SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pOid1", "pOid2", "nSubIds"]),
        #
        'SnmpUtilOidCmp': SimTypeFunction([SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pOid1", "pOid2"]),
        #
        'SnmpUtilOidFree': SimTypeFunction([SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pOid"]),
        #
        'SnmpUtilOctetsCmp': SimTypeFunction([SimTypePointer(SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pOctets1", "pOctets2"]),
        #
        'SnmpUtilOctetsNCmp': SimTypeFunction([SimTypePointer(SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pOctets1", "pOctets2", "nChars"]),
        #
        'SnmpUtilOctetsCpy': SimTypeFunction([SimTypePointer(SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pOctetsDst", "pOctetsSrc"]),
        #
        'SnmpUtilOctetsFree': SimTypeFunction([SimTypePointer(SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pOctets"]),
        #
        'SnmpUtilAsnAnyCpy': SimTypeFunction([SimTypePointer(SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAnyDst", "pAnySrc"]),
        #
        'SnmpUtilAsnAnyFree': SimTypeFunction([SimTypePointer(SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pAny"]),
        #
        'SnmpUtilVarBindCpy': SimTypeFunction([SimTypePointer(SimStruct({"name": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "value": SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None)}, name="SnmpVarBind", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"name": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "value": SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None)}, name="SnmpVarBind", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pVbDst", "pVbSrc"]),
        #
        'SnmpUtilVarBindFree': SimTypeFunction([SimTypePointer(SimStruct({"name": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "value": SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None)}, name="SnmpVarBind", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pVb"]),
        #
        'SnmpUtilVarBindListCpy': SimTypeFunction([SimTypePointer(SimStruct({"list": SimTypePointer(SimStruct({"name": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "value": SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None)}, name="SnmpVarBind", pack=False, align=None), offset=0), "len": SimTypeInt(signed=False, label="UInt32")}, name="SnmpVarBindList", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"list": SimTypePointer(SimStruct({"name": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "value": SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None)}, name="SnmpVarBind", pack=False, align=None), offset=0), "len": SimTypeInt(signed=False, label="UInt32")}, name="SnmpVarBindList", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pVblDst", "pVblSrc"]),
        #
        'SnmpUtilVarBindListFree': SimTypeFunction([SimTypePointer(SimStruct({"list": SimTypePointer(SimStruct({"name": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "value": SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None)}, name="SnmpVarBind", pack=False, align=None), offset=0), "len": SimTypeInt(signed=False, label="UInt32")}, name="SnmpVarBindList", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pVbl"]),
        #
        'SnmpUtilMemFree': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["pMem"]),
        #
        'SnmpUtilMemAlloc': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["nBytes"]),
        #
        'SnmpUtilMemReAlloc': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["pMem", "nBytes"]),
        #
        'SnmpUtilOidToA': SimTypeFunction([SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0)], SimTypePointer(SimTypeChar(label="Byte"), offset=0), arg_names=["Oid"]),
        #
        'SnmpUtilIdsToA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeChar(label="Byte"), offset=0), arg_names=["Ids", "IdLength"]),
        #
        'SnmpUtilPrintOid': SimTypeFunction([SimTypePointer(SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["Oid"]),
        #
        'SnmpUtilPrintAsnAny': SimTypeFunction([SimTypePointer(SimStruct({"asnType": SimTypeChar(label="Byte"), "asnValue": SimUnion({"number": SimTypeInt(signed=True, label="Int32"), "unsigned32": SimTypeInt(signed=False, label="UInt32"), "counter64": SimTypeBottom(label="ULARGE_INTEGER"), "string": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "bits": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "object": SimStruct({"idLength": SimTypeInt(signed=False, label="UInt32"), "ids": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="AsnObjectIdentifier", pack=False, align=None), "sequence": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "address": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None), "counter": SimTypeInt(signed=False, label="UInt32"), "gauge": SimTypeInt(signed=False, label="UInt32"), "ticks": SimTypeInt(signed=False, label="UInt32"), "arbitrary": SimStruct({"stream": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "length": SimTypeInt(signed=False, label="UInt32"), "dynamic": SimTypeInt(signed=True, label="Int32")}, name="AsnOctetString", pack=False, align=None)}, name="<anon>", label="None")}, name="AsnAny", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pAny"]),
        #
        'SnmpSvcGetUptime': SimTypeFunction([], SimTypeInt(signed=False, label="UInt32")),
        #
        'SnmpSvcSetLogLevel': SimTypeFunction([SimTypeInt(signed=False, label="SNMP_LOG")], SimTypeBottom(label="Void"), arg_names=["nLogLevel"]),
        #
        'SnmpSvcSetLogType': SimTypeFunction([SimTypeInt(signed=False, label="SNMP_OUTPUT_LOG_TYPE")], SimTypeBottom(label="Void"), arg_names=["nLogType"]),
        #
        'SnmpUtilDbgPrint': SimTypeFunction([SimTypeInt(signed=False, label="SNMP_LOG"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["nLogLevel", "szFormat"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("dciman32.dll")
prototypes = \
    {
        #
        'DCIOpenProvider': SimTypeFunction([], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)),
        #
        'DCICloseProvider': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["hdc"]),
        #
        'DCICreatePrimary': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "lplpSurface"]),
        #
        'DCICreateOffscreen': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"dciInfo": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), "Draw": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetClipList": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetDestination": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCIOFFSCREEN", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "dwCompression", "dwRedMask", "dwGreenMask", "dwBlueMask", "dwWidth", "dwHeight", "dwDCICaps", "dwBitCount", "lplpSurface"]),
        #
        'DCICreateOverlay': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dciInfo": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), "dwChromakeyValue": SimTypeInt(signed=False, label="UInt32"), "dwChromakeyMask": SimTypeInt(signed=False, label="UInt32")}, name="DCIOVERLAY", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "lpOffscreenSurf", "lplpSurface"]),
        #
        'DCIEnum': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "lprDst", "lprSrc", "lpFnCallback", "lpContext"]),
        #
        'DCISetSrcDestClip': SimTypeFunction([SimTypePointer(SimStruct({"dciInfo": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), "Draw": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetClipList": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetDestination": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCIOFFSCREEN", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"rdh": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "iType": SimTypeInt(signed=False, label="UInt32"), "nCount": SimTypeInt(signed=False, label="UInt32"), "nRgnSize": SimTypeInt(signed=False, label="UInt32"), "rcBound": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None)}, name="RGNDATAHEADER", pack=False, align=None), "Buffer": SimTypePointer(SimTypeBottom(label="CHAR"), offset=0)}, name="RGNDATA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pdci", "srcrc", "destrc", "prd"]),
        #
        'WinWatchOpen': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwnd"]),
        #
        'WinWatchClose': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["hWW"]),
        #
        'WinWatchGetClipList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"rdh": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "iType": SimTypeInt(signed=False, label="UInt32"), "nCount": SimTypeInt(signed=False, label="UInt32"), "nRgnSize": SimTypeInt(signed=False, label="UInt32"), "rcBound": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None)}, name="RGNDATAHEADER", pack=False, align=None), "Buffer": SimTypePointer(SimTypeBottom(label="CHAR"), offset=0)}, name="RGNDATA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hWW", "prc", "size", "prd"]),
        #
        'WinWatchDidStatusChange': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWW"]),
        #
        'GetWindowRegionData': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"rdh": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "iType": SimTypeInt(signed=False, label="UInt32"), "nCount": SimTypeInt(signed=False, label="UInt32"), "nRgnSize": SimTypeInt(signed=False, label="UInt32"), "rcBound": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None)}, name="RGNDATAHEADER", pack=False, align=None), "Buffer": SimTypePointer(SimTypeBottom(label="CHAR"), offset=0)}, name="RGNDATA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hwnd", "size", "prd"]),
        #
        'GetDCRegionData': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"rdh": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "iType": SimTypeInt(signed=False, label="UInt32"), "nCount": SimTypeInt(signed=False, label="UInt32"), "nRgnSize": SimTypeInt(signed=False, label="UInt32"), "rcBound": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None)}, name="RGNDATAHEADER", pack=False, align=None), "Buffer": SimTypePointer(SimTypeBottom(label="CHAR"), offset=0)}, name="RGNDATA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hdc", "size", "prd"]),
        #
        'WinWatchNotify': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["hww", "hwnd", "code", "lParam"]), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWW", "NotifyCallback", "NotifyParam"]),
        #
        'DCIEndAccess': SimTypeFunction([SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pdci"]),
        #
        'DCIBeginAccess': SimTypeFunction([SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pdci", "x", "y", "dx", "dy"]),
        #
        'DCIDestroy': SimTypeFunction([SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pdci"]),
        #
        'DCIDraw': SimTypeFunction([SimTypePointer(SimStruct({"dciInfo": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), "Draw": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetClipList": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetDestination": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCIOFFSCREEN", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pdci"]),
        #
        'DCISetClipList': SimTypeFunction([SimTypePointer(SimStruct({"dciInfo": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), "Draw": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetClipList": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetDestination": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCIOFFSCREEN", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"rdh": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "iType": SimTypeInt(signed=False, label="UInt32"), "nCount": SimTypeInt(signed=False, label="UInt32"), "nRgnSize": SimTypeInt(signed=False, label="UInt32"), "rcBound": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None)}, name="RGNDATAHEADER", pack=False, align=None), "Buffer": SimTypePointer(SimTypeBottom(label="CHAR"), offset=0)}, name="RGNDATA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pdci", "prd"]),
        #
        'DCISetDestination': SimTypeFunction([SimTypePointer(SimStruct({"dciInfo": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDCICaps": SimTypeInt(signed=False, label="UInt32"), "dwCompression": SimTypeInt(signed=False, label="UInt32"), "dwMask": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 3), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "lStride": SimTypeInt(signed=True, label="Int32"), "dwBitCount": SimTypeInt(signed=False, label="UInt32"), "dwOffSurface": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "wSelSurface": SimTypeShort(signed=False, label="UInt16"), "wReserved": SimTypeShort(signed=False, label="UInt16"), "dwReserved1": SimTypeInt(signed=False, label="UInt32"), "dwReserved2": SimTypeInt(signed=False, label="UInt32"), "dwReserved3": SimTypeInt(signed=False, label="UInt32"), "BeginAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "EndAccess": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestroySurface": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCISURFACEINFO", pack=False, align=None), "Draw": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetClipList": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetDestination": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="DCIOFFSCREEN", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pdci", "dst", "src"]),
    }

lib.set_prototypes(prototypes)



from typing import Set, Optional, Literal, Type, TypeVar, Union, Generic
from dataclasses import dataclass
import logging

from angr.knowledge_plugins.variables.variable_manager import VariableManagerInternal
from angr.sim_variable import SimTemporaryVariable
from angr.sim_variable import SimMemoryVariable
from angr.sim_variable import SimStackVariable
from angr.sim_variable import SimRegisterVariable
from angr.misc.ux import once

from ...engines.light import SpOffset
from ...code_location import CodeLocation, ExternalCodeLocation
from .atoms import Atom, MemoryLocation, Register, Tmp, AtomKind, atom_kind_mapping
from .tag import Tag
from ...sim_variable import SimVariable

log = logging.getLogger(__name__)


@dataclass
class DefinitionMatchPredicate:
    """
    A dataclass indicating several facts which much all must match in order for a definition to match. Largely an
    internal class; don't worry about this.
    """

    kind: Union[AtomKind, Type[Atom], None] = None
    bbl_addr: Optional[int] = None
    ins_addr: Optional[int] = None
    variable: Optional[SimVariable] = None
    variable_manager: Union[VariableManagerInternal, None, Literal[False]] = None
    stack_offset: Optional[int] = None
    reg_name: Optional[Union[str, int]] = None
    heap_offset: Optional[int] = None
    global_addr: Optional[int] = None
    tmp_idx: Optional[int] = None
    const_val: Optional[int] = None
    extern: Optional[bool] = None

    @staticmethod
    def construct(predicate: Optional["DefinitionMatchPredicate"] = None, **kwargs) -> "DefinitionMatchPredicate":
        if predicate is None:
            predicate = DefinitionMatchPredicate(**kwargs)
            predicate.normalize()
        return predicate

    def normalize(self):
        if self.variable is not None:
            if isinstance(self.variable, SimRegisterVariable):
                self.reg_name = self.variable.reg
            elif isinstance(self.variable, SimStackVariable):
                if self.variable.base != "bp":
                    log.warning("Cannot match against variables with %s base (need bp)", self.variable.base)
                else:
                    self.stack_offset = self.variable.offset
            elif isinstance(self.variable, SimMemoryVariable):
                # TODO region
                if isinstance(self.variable.addr, int):
                    self.global_addr = self.variable.addr
                else:
                    log.warning(
                        "Cannot match against memory variable with %s addr (need int)",
                        type(self.variable.addr).__name__,
                    )
            elif isinstance(self.variable, SimTemporaryVariable):
                self.tmp_idx = self.variable.tmp_id
            else:
                log.warning(
                    "Cannot match against definition to %s (need reg, stack, mem, or tmp)", type(self.variable).__name__
                )

        if self.reg_name is not None:
            self.kind = AtomKind.REGISTER
        elif self.stack_offset is not None or self.heap_offset is not None or self.global_addr is not None:
            self.kind = AtomKind.MEMORY
        elif self.const_val is not None:
            self.kind = AtomKind.CONSTANT
        elif self.tmp_idx is not None:
            self.kind = AtomKind.TMP

    def matches(self, defn: "Definition") -> bool:
        if self.variable is not None:
            if self.variable_manager is False:
                pass
            elif self.variable_manager is not None:
                if not self.variable_manager.is_variable_used_at(
                    self.variable, (defn.codeloc.bbl_addr, defn.codeloc.stmt_idx)
                ):
                    return False
            elif once("definition_matches_no_variable_manager"):
                log.warning(
                    "Cannot match definitions to variables on the basis of locations without a variable manager."
                )
                log.warning("Pass variable_manager=False to acknowledge this explicitly.")
        if self.bbl_addr is not None and defn.codeloc.block_addr != self.bbl_addr:
            return False
        if self.ins_addr is not None and defn.codeloc.ins_addr != self.ins_addr:
            return False
        if self.extern is not None and isinstance(defn.codeloc, ExternalCodeLocation) != self.extern:
            return False

        if self.kind is not None:
            if not isinstance(self.kind, type):
                self.kind = atom_kind_mapping[self.kind]
            if not isinstance(defn.atom, self.kind):
                return False

        if isinstance(defn.atom, Register):
            if self.reg_name is not None:
                if isinstance(self.reg_name, int):
                    if not defn.atom.reg_offset <= self.reg_name < defn.atom.reg_offset + defn.atom.size:
                        return False
                elif isinstance(self.reg_name, str):
                    if defn.atom.arch is not None:
                        if self.reg_name != defn.atom.name:
                            return False
                    else:
                        log.warning(
                            "Attempting to match by register name against a definition which does not have an arch"
                        )
                        return False
                else:
                    raise TypeError(self.reg_name)
        elif isinstance(defn.atom, MemoryLocation):
            if self.stack_offset is not None:
                if (
                    not isinstance(defn.atom.addr, SpOffset)
                    or defn.atom.addr.base != "sp"  # TODO???????
                    or defn.atom.addr.offset != self.stack_offset
                ):
                    return False
        elif isinstance(defn.atom, Tmp):
            if self.tmp_idx is not None and self.tmp_idx != defn.atom.tmp_idx:
                return False

        return True


A = TypeVar("A", bound=Atom)


class Definition(Generic[A]):
    """
    An atom definition.

    :ivar atom:     The atom being defined.
    :ivar codeloc:  Where this definition is created in the original binary code.
    :ivar dummy:    Tell whether the definition should be considered dummy or not. During simplification by AILment,
                    definitions marked as dummy will not be removed.
    :ivar tags:     A set of tags containing information about the definition gathered during analyses.
    """

    __slots__ = (
        "atom",
        "codeloc",
        "dummy",
        "tags",
        "_hash",
    )

    def __init__(self, atom: A, codeloc: CodeLocation, dummy: bool = False, tags: Optional[Set[Tag]] = None):
        self.atom: A = atom
        self.codeloc: CodeLocation = codeloc
        self.dummy: bool = dummy
        self.tags = tags or set()
        self._hash = None

    def __eq__(self, other):
        return self.atom == other.atom and self.codeloc == other.codeloc

    def __repr__(self):
        if not self.tags:
            return "<Definition {{Atom:{}, Codeloc:{}}}{}>".format(
                self.atom, self.codeloc, "" if not self.dummy else "dummy"
            )
        else:
            return "<Definition {{Tags:{}, Atom:{}, Codeloc:{}}}{}>".format(
                repr(self.tags), self.atom, self.codeloc, "" if not self.dummy else " dummy"
            )

    def __str__(self):
        pretty_tags = "\n".join([str(tag) for tag in self.tags])
        return f"Definition:\n" f"Atom: {self.atom}\n" f"CodeLoc: {self.codeloc}\n" f"Tags: {pretty_tags}"

    def __hash__(self):
        if self._hash is None:
            self._hash = hash((self.atom, self.codeloc))
        return self._hash

    @property
    def offset(self) -> int:
        if isinstance(self.atom, Register):
            return self.atom.reg_offset
        elif isinstance(self.atom, MemoryLocation):
            if isinstance(self.atom.addr, SpOffset):
                return self.atom.addr.offset
            else:
                return self.atom.addr
        else:
            raise ValueError("Unsupported operation offset on %s." % type(self.atom))

    @property
    def size(self) -> int:
        if isinstance(self.atom, Register):
            return self.atom.size
        elif isinstance(self.atom, MemoryLocation):
            return self.atom.bits // 8
        else:
            raise ValueError("Unsupported operation size on %s." % type(self.atom))

    def matches(self, **kwargs) -> bool:
        """
        Return whether this definition has certain characteristics.

        """
        return DefinitionMatchPredicate.construct(**kwargs).matches(self)



import argparse

from angr.analyses.decompiler.structuring import STRUCTURER_CLASSES
from angr.analyses.decompiler.utils import decompile_functions


class COMMANDS:
    """
    The commands that the angr CLI supports.
    """

    DECOMPILE = "decompile"
    ALL_COMMANDS = [DECOMPILE]


def main():
    parser = argparse.ArgumentParser(description="The angr CLI allows you to decompile and analyze binaries.")
    parser.add_argument(
        "command",
        help="""
        The analysis type to run on the binary. All analysis is output to stdout.""",
        choices=COMMANDS.ALL_COMMANDS,
    )
    parser.add_argument("binary", help="The path to the binary to analyze.")
    parser.add_argument(
        "--functions",
        help="""
        The functions to analyze under the current command. Functions can either be expressed as names found in the
        symbols of the binary or as addresses like: 0x401000.""",
        nargs="+",
    )
    parser.add_argument(
        "--catch-exceptions",
        help="""
        Catch exceptions during analysis. The scope of error handling may depend on the command used for analysis.
        If multiple functions are specified for analysis, each function will be handled individually.""",
        action="store_true",
        default=False,
    )
    # decompilation-specific arguments
    parser.add_argument(
        "--structurer",
        help="The structuring algorithm to use for decompilation.",
        choices=STRUCTURER_CLASSES.keys(),
        default="phoenix",
    )

    args = parser.parse_args()
    if args.command == COMMANDS.DECOMPILE:
        decompilation = decompile_functions(
            args.binary, functions=args.functions, structurer=args.structurer, catch_errors=args.catch_exceptions
        )
        print(decompilation)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from .graph_region import GraphRegion


class RegionWalker:
    """
    A simple traverser class that walks GraphRegion instances.
    """

    def __init__(self):
        self._parent_region = None
        self._current_region = None

    def walk(self, region: GraphRegion):
        for node in region.graph.nodes():
            if isinstance(node, GraphRegion):
                self._parent_region = node
                self.walk(node)
                self._parent_region = None
            else:
                self.walk_node(region, node)

    def walk_node(self, region, node):  # pylint:disable=no-self-use,unused-argument
        raise NotImplementedError("Please override this method with your own logic")



from ailment.expression import BinaryOp, Convert, Const, ITE

from .base import PeepholeOptimizationExprBase

_MASKS = {
    1: 1,
    8: 0xFF,
    16: 0xFFFF,
    32: 0xFFFFFFFF,
    64: 0xFFFFFFFF_FFFFFFFF,
}


class RemoveRedundantBitmasks(PeepholeOptimizationExprBase):
    """
    Remove redundant bitmasking operations.
    """

    __slots__ = ()

    NAME = "Remove redundant bitmasks"
    expr_classes = (BinaryOp,)

    def optimize(self, expr: BinaryOp, **kwargs):
        # And(expr, full_N_bitmask) ==> expr
        # And(Conv(1->N, expr), bitmask) ==> Conv(1->N, expr)
        # And(Conv(1->N, bool_expr), bitmask) ==> Conv(1->N, bool_expr)
        # And(ITE(?, const_expr, const_expr), bitmask) ==> ITE(?, const_expr, const_expr)
        if expr.op == "And" and isinstance(expr.operands[1], Const):
            inner_expr = expr.operands[0]
            if expr.operands[1].value == _MASKS.get(inner_expr.bits, None):
                return inner_expr
            if isinstance(inner_expr, Convert) and self.is_bool_expr(inner_expr.operand):
                # useless masking
                return inner_expr
            if (
                isinstance(expr.operands[0], ITE)
                and isinstance(expr.operands[0].iftrue, Const)
                and isinstance(expr.operands[0].iffalse, Const)
            ):
                # is the masking unnecessary?
                mask = expr.operands[1].value
                ite = expr.operands[0]
                if mask == 0xFF and ite.iftrue.value <= 0xFF and ite.iffalse.value <= 0xFF:
                    # yes!
                    return ite

        return None



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("api-ms-win-core-winrt-l1-1-0.dll")
prototypes = \
    {
        #
        'RoInitialize': SimTypeFunction([SimTypeInt(signed=False, label="RO_INIT_TYPE")], SimTypeInt(signed=True, label="Int32"), arg_names=["initType"]),
        #
        'RoUninitialize': SimTypeFunction([], SimTypeBottom(label="Void")),
        #
        'RoActivateInstance': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="IInspectable"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["activatableClassId", "instance"]),
        #
        'RoRegisterActivationFactories': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["activatableClassIds", "activationFactoryCallbacks", "count", "cookie"]),
        #
        'RoRevokeActivationFactories': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["cookie"]),
        #
        'RoGetActivationFactory': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["activatableClassId", "iid", "factory"]),
        #
        'RoRegisterForApartmentShutdown': SimTypeFunction([SimTypeBottom(label="IApartmentShutdown"), SimTypePointer(SimTypeLongLong(signed=False, label="UInt64"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["callbackObject", "apartmentIdentifier", "regCookie"]),
        #
        'RoUnregisterForApartmentShutdown': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["regCookie"]),
        #
        'RoGetApartmentIdentifier': SimTypeFunction([SimTypePointer(SimTypeLongLong(signed=False, label="UInt64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["apartmentIdentifier"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=unused-argument,no-self-use
from collections import OrderedDict

import claripy
import ailment
from ailment.expression import negate

from .sequence_walker import SequenceWalker
from .structuring.structurer_nodes import (
    MultiNode,
    SequenceNode,
    CodeNode,
    ConditionNode,
    SwitchCaseNode,
    ConditionalBreakNode,
    BreakNode,
    LoopNode,
    CascadingConditionNode,
    ContinueNode,
)
from .condition_processor import ConditionProcessor


class EmptyNodeRemover:
    """
    Rewrites a node and its children to remove empty nodes.

    The following optimizations are performed at the same time:
    - Convert if (A) { } else { ... } to if(!A) { ... } else { }

    :ivar _claripy_ast_conditions:  True if all node conditions are claripy ASTs. False if all node conditions are AIL
                                    expressions.
    """

    def __init__(self, node, claripy_ast_conditions: bool = True):
        self.root = node
        self._claripy_ast_conditions = claripy_ast_conditions

        self.removed_sequences = []
        self.replaced_sequences = {}

        handlers = {
            SequenceNode: self._handle_Sequence,
            CodeNode: self._handle_Code,
            ConditionNode: self._handle_Condition,
            CascadingConditionNode: self._handle_CascadingCondition,
            SwitchCaseNode: self._handle_SwitchCase,
            LoopNode: self._handle_Loop,
            ContinueNode: self._handle_Continue,
            MultiNode: self._handle_MultiNode,
            BreakNode: self._handle_Default,
            ConditionalBreakNode: self._handle_Default,
            ailment.Block: self._handle_Block,
        }
        self._walker = SequenceWalker(handlers=handlers)
        r = self._walker.walk(self.root)
        if r is None:
            self.result = SequenceNode(None, nodes=[])
        else:
            # Make sure it's still a sequence node
            if not isinstance(r, SequenceNode):
                r = SequenceNode(node.addr, nodes=[r])
            self.result = r

    #
    # Handlers
    #

    def _handle_Sequence(self, node, **kwargs):
        new_nodes = []
        for node_ in node.nodes:
            new_node = self._walker._handle(node_)
            if new_node is not None:
                if isinstance(new_node, SequenceNode):
                    new_nodes.extend(new_node.nodes)
                else:
                    new_nodes.append(new_node)

        if not new_nodes:
            self.removed_sequences.append(node)
            return None
        if len(new_nodes) == 1:
            # Remove the unnecessary sequence node
            self.replaced_sequences[node] = new_nodes[0]
            return new_nodes[0]

        sn = SequenceNode(node.addr, nodes=new_nodes)
        self.replaced_sequences[node] = sn
        return sn

    def _handle_MultiNode(self, node: MultiNode, **kwargs):
        new_nodes = []
        for node_ in node.nodes:
            new_node = self._walker._handle(node_)
            if new_node is not None:
                if isinstance(new_node, MultiNode):
                    new_nodes.extend(new_node.nodes)
                else:
                    new_nodes.append(new_node)

        if not new_nodes:
            return None
        if len(new_nodes) == 1:
            # Remove the unnecessary MultiNode
            return new_nodes[0]
        return MultiNode(new_nodes)

    def _handle_Code(self, node, **kwargs):
        inner_node = self._walker._handle(node.node)
        if inner_node is None:
            return None
        if (
            self._claripy_ast_conditions
            and node.reaching_condition is not None
            and claripy.is_true(node.reaching_condition)
        ):
            # Remove the unnecessary CodeNode
            return inner_node
        if self._claripy_ast_conditions and isinstance(inner_node, CodeNode):
            # unpack the codenode so we don't have directly nested CodeNodes
            return CodeNode(inner_node.node, claripy.And(node.reaching_condition, inner_node.reaching_condition))
        return CodeNode(inner_node, node.reaching_condition)

    def _handle_Condition(self, node, **kwargs):
        true_node = self._walker._handle(node.true_node)
        false_node = self._walker._handle(node.false_node)

        if true_node is None and false_node is None:
            # empty node
            return None
        if true_node is None and false_node is not None:
            # swap them
            return ConditionNode(
                node.addr,
                node.reaching_condition,
                (
                    ConditionProcessor.simplify_condition(claripy.Not(node.condition))
                    if self._claripy_ast_conditions
                    else negate(node.condition)
                ),
                false_node,
                false_node=None,
            )
        if (
            self._claripy_ast_conditions
            and claripy.is_true(node.condition)
            and node.true_node is not None
            and node.false_node is None
        ):
            return node.true_node
        return ConditionNode(node.addr, node.reaching_condition, node.condition, true_node, false_node=false_node)

    def _handle_CascadingCondition(self, node: CascadingConditionNode, **kwargs):
        new_else_node = None if node.else_node is None else self._walker._handle(node.else_node)

        new_cond_and_nodes = []
        for cond, child_node in node.condition_and_nodes:
            new_node = self._walker._handle(child_node)
            if new_node is not None:
                new_cond_and_nodes.append((cond, new_node))
            else:
                if new_else_node is not None:
                    # do not allow any empty condition nodes, otherwise the condition for the else node will be wrong
                    new_cond_and_nodes.append((cond, child_node))

        if not new_cond_and_nodes and new_else_node is None:
            # empty node
            return None
        return CascadingConditionNode(node.addr, new_cond_and_nodes, else_node=new_else_node)

    def _handle_Loop(self, node: LoopNode, **kwargs):
        new_seq = self._walker._handle(node.sequence_node)

        if (
            new_seq is None
            and node.sort == "while"
            and (node.condition is None or (isinstance(node.condition, ailment.Const) and node.condition.value == 0))
        ):
            return None

        result = node.copy()
        result.sequence_node = new_seq
        return result

    def _handle_Continue(self, node: ContinueNode, **kwargs):
        return node

    def _handle_SwitchCase(self, node, **kwargs):
        new_cases = OrderedDict()

        for idx, case in node.cases.items():
            new_case = self._walker._handle(case)
            if new_case is not None:
                new_cases[idx] = new_case

        new_default_node = self._walker._handle(node.default_node)

        if not new_cases and new_default_node is None:
            return None

        return SwitchCaseNode(node.switch_expr, new_cases, new_default_node, addr=node.addr)

    @staticmethod
    def _handle_Default(node, **kwargs):
        return node

    @staticmethod
    def _handle_Block(block, **kwargs):
        if not block.statements:
            return None
        return block



import random

from ..func import Func, TestData


def rand_str(length, byte_list=None):
    if byte_list is None:
        return "".join(chr(random.randint(0, 255)) for _ in range(length))
    return "".join(random.choice(byte_list) for _ in range(length))


class memset(Func):
    non_null = [chr(i) for i in range(1, 256)]

    def __init__(self):
        super().__init__()  # pylint disable=useless-super-delegation

    def get_name(self):
        return "memset"

    def num_args(self):
        return 3

    def args(self):  # pylint disable=no-self-use
        return ["buf", "char", "size"]

    def can_call_other_funcs(self):
        return False

    def gen_input_output_pair(self):
        # TODO we don't check the return val
        set_len = random.randint(1, 40)
        char = random.randint(0, 255)
        result_buf = rand_str(set_len + 5)
        test_input = [result_buf, char, set_len]
        test_output = [chr(char) * set_len + result_buf[-5:], None, None]
        max_steps = 20
        return_val = None
        test = TestData(test_input, test_output, return_val, max_steps)
        return test

    def pre_test(self, func, runner):
        return True



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("api-ms-win-core-winrt-error-l1-1-0.dll")
prototypes = \
    {
        #
        'RoGetErrorReportingFlags': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pflags"]),
        #
        'RoSetErrorReportingFlags': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["flags"]),
        #
        'RoResolveRestrictedErrorInfoReference': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="IRestrictedErrorInfo"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["reference", "ppRestrictedErrorInfo"]),
        #
        'SetRestrictedErrorInfo': SimTypeFunction([SimTypeBottom(label="IRestrictedErrorInfo")], SimTypeInt(signed=True, label="Int32"), arg_names=["pRestrictedErrorInfo"]),
        #
        'GetRestrictedErrorInfo': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IRestrictedErrorInfo"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppRestrictedErrorInfo"]),
        #
        'RoOriginateErrorW': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["error", "cchMax", "message"]),
        #
        'RoOriginateError': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["error", "message"]),
        #
        'RoTransformErrorW': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["oldError", "newError", "cchMax", "message"]),
        #
        'RoTransformError': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["oldError", "newError", "message"]),
        #
        'RoCaptureErrorContext': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hr"]),
        #
        'RoFailFastWithErrorContext': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeBottom(label="Void"), arg_names=["hrError"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=unused-argument
from typing import Optional, Dict, Set, Tuple, Generator, TYPE_CHECKING
from enum import Enum

import networkx  # pylint:disable=unused-import
import ailment

from angr.analyses.decompiler import RegionIdentifier
from angr.analyses.decompiler.goto_manager import GotoManager
from angr.analyses.decompiler.structuring import RecursiveStructurer, PhoenixStructurer
from angr.analyses.decompiler.utils import add_labels

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function


class MultipleBlocksException(Exception):
    """
    An exception that is raised in _get_block() where multiple blocks satisfy the criteria but only one block was
    requested.
    """


class OptimizationPassStage(Enum):
    """
    Enums about optimization pass stages.

    Note that the region identification pass (RegionIdentifier) may modify existing AIL blocks *without updating the
    topology of the original AIL graph*. For example, loop successor refinement may modify create a new AIL block with
    an artificial address, and alter existing jump targets of jump statements and conditional jump statements to point
    to this new block. However, loop successor refinement does not update the topology of the original AIL graph, which
    means this new AIL block does not exist in the original AIL graph. As a result, until this behavior of
    RegionIdentifier changes in the future, DURING_REGION_IDENTIFICATION optimization passes should not modify existing
    jump targets.
    """

    AFTER_AIL_GRAPH_CREATION = 0
    AFTER_SINGLE_BLOCK_SIMPLIFICATION = 1
    AFTER_MAKING_CALLSITES = 2
    AFTER_GLOBAL_SIMPLIFICATION = 3
    AFTER_VARIABLE_RECOVERY = 4
    BEFORE_REGION_IDENTIFICATION = 5
    DURING_REGION_IDENTIFICATION = 6
    AFTER_STRUCTURING = 7


class BaseOptimizationPass:
    """
    The base class for any optimization pass.
    """

    ARCHES = []  # strings of supported architectures
    PLATFORMS = []  # strings of supported platforms. Can be one of the following: "win32", "linux"
    STAGE: int = None  # Specifies when this optimization pass should be executed
    STRUCTURING: Optional[str] = (
        None  # specifies if this optimization pass is specific to a certain structuring algorithm
    )
    NAME = "N/A"
    DESCRIPTION = "N/A"

    def __init__(self, func):
        self._func: "Function" = func

    @property
    def project(self):
        return self._func.project

    @property
    def kb(self):
        return self.project.kb

    def analyze(self):
        ret, cache = self._check()
        if ret:
            self._analyze(cache=cache)

    def _check(self):
        """
        Check if this optimization applies to this function.

        :returns: a tuple of (does_apply, cache) where cache is a way to pass
                  information to _analyze so it does not have to be recalculated
        """
        raise NotImplementedError()

    def _analyze(self, cache=None):
        """
        Run the analysis.

        :param cache: information passed from _check so it does not have to be
                      recalculated
        :returns: None
        """
        raise NotImplementedError()


class OptimizationPass(BaseOptimizationPass):
    """
    The base class for any function-level graph optimization pass.
    """

    def __init__(
        self,
        func,
        blocks_by_addr=None,
        blocks_by_addr_and_idx=None,
        graph=None,
        variable_kb=None,
        region_identifier=None,
        reaching_definitions=None,
        **kwargs,
    ):
        super().__init__(func)
        # self._blocks is just a cache
        self._blocks_by_addr: Dict[int, Set[ailment.Block]] = blocks_by_addr
        self._blocks_by_addr_and_idx: Dict[Tuple[int, Optional[int]], ailment.Block] = blocks_by_addr_and_idx
        self._graph: Optional[networkx.DiGraph] = graph
        self._variable_kb = variable_kb
        self._ri = region_identifier
        self._rd = reaching_definitions
        self._new_block_addrs = set()

        # output
        self.out_graph: Optional[networkx.DiGraph] = None

    @property
    def blocks_by_addr(self) -> Dict[int, Set[ailment.Block]]:
        return self._blocks_by_addr

    @property
    def blocks_by_addr_and_idx(self) -> Dict[Tuple[int, Optional[int]], ailment.Block]:
        return self._blocks_by_addr_and_idx

    #
    # Util methods
    #

    def new_block_addr(self) -> int:
        """
        Return a block address that does not conflict with any existing blocks.

        :return:    The block address.
        """
        if self._new_block_addrs:
            new_addr = max(self._new_block_addrs) + 1
        else:
            new_addr = max(self.blocks_by_addr) + 2048
        self._new_block_addrs.add(new_addr)
        return new_addr

    def _get_block(self, addr, idx=None) -> Optional[ailment.Block]:
        if not self._blocks_by_addr:
            return None
        else:
            if idx is None:
                blocks = self._blocks_by_addr.get(addr, None)
            else:
                blocks = [self._blocks_by_addr_and_idx.get((addr, idx), None)]
            if not blocks:
                return None
            if len(blocks) == 1:
                return next(iter(blocks))
            raise MultipleBlocksException(
                "There are %d blocks at address %#x.%s but only one is requested." % (len(blocks), addr, idx)
            )

    def _get_blocks(self, addr, idx=None) -> Generator[ailment.Block, None, None]:
        if not self._blocks_by_addr:
            return
        else:
            if idx is None:
                blocks = self._blocks_by_addr.get(addr, None)
                if blocks is not None:
                    yield from blocks
            else:
                block = self._blocks_by_addr_and_idx.get((addr, idx), None)
                if block is not None:
                    yield block

    def _update_block(self, old_block, new_block):
        if self.out_graph is None:
            self.out_graph = self._graph  # we do not make copy here for performance reasons. we can change it if needed

        if old_block not in self.out_graph:
            return

        in_edges = list(self.out_graph.in_edges(old_block, data=True))
        out_edges = list(self.out_graph.out_edges(old_block, data=True))

        self._remove_block(old_block)
        self.out_graph.add_node(new_block)
        self._blocks_by_addr[new_block.addr].add(new_block)
        self._blocks_by_addr_and_idx[(new_block.addr, new_block.idx)] = new_block

        for src, _, data in in_edges:
            if src is old_block:
                src = new_block
            self.out_graph.add_edge(src, new_block, **data)

        for _, dst, data in out_edges:
            if dst is old_block:
                dst = new_block
            self.out_graph.add_edge(new_block, dst, **data)

    def _remove_block(self, block):
        if self.out_graph is None:
            self.out_graph = self._graph

        if block in self.out_graph:
            self.out_graph.remove_node(block)

        if block.addr in self._blocks_by_addr and block in self._blocks_by_addr[block.addr]:
            self._blocks_by_addr[block.addr].remove(block)
            del self._blocks_by_addr_and_idx[(block.addr, block.idx)]

    @staticmethod
    def _is_add(expr):
        return isinstance(expr, ailment.Expr.BinaryOp) and expr.op == "Add"

    @staticmethod
    def _is_sub(expr):
        return isinstance(expr, ailment.Expr.BinaryOp) and expr.op == "Sub"


class SequenceOptimizationPass(BaseOptimizationPass):
    """
    The base class for any sequence node optimization pass.
    """

    ARCHES = []  # strings of supported architectures
    PLATFORMS = []  # strings of supported platforms. Can be one of the following: "win32", "linux"
    STAGE: int = None  # Specifies when this optimization pass should be executed

    def __init__(self, func, seq=None, **kwargs):
        super().__init__(func)
        self.seq = seq
        self.out_seq = None


class StructuringOptimizationPass(OptimizationPass):
    """
    The base class for any optimization pass that requires structuring. Optimization passes that inherit from this class
    should directly depend on structuring artifacts, such as regions and gotos. Otherwise, they should use
    OptimizationPass. This is the heaviest (computation time) optimization pass class.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.DURING_REGION_IDENTIFICATION

    def __init__(
        self,
        func,
        prevent_new_gotos=True,
        strictly_less_gotos=False,
        recover_structure_fails=True,
        max_opt_iters=1,
        simplify_ail=True,
        **kwargs,
    ):
        super().__init__(func, **kwargs)
        self._prevent_new_gotos = prevent_new_gotos
        self._strictly_less_gotos = strictly_less_gotos
        self._recover_structure_fails = recover_structure_fails
        self._max_opt_iters = max_opt_iters
        self._simplify_ail = simplify_ail

        self._goto_manager: Optional[GotoManager] = None
        self._prev_graph: Optional[networkx.DiGraph] = None

    def _analyze(self, cache=None) -> bool:
        raise NotImplementedError()

    def analyze(self):
        """
        Wrapper for _analyze() that verifies the graph is structurable before and after the optimization.
        """
        if not self._graph_is_structurable(self._graph):
            return

        initial_gotos = self._goto_manager.gotos.copy()
        # replace the normal check in OptimizationPass.analyze()
        ret, cache = self._check()
        if not ret:
            return

        # setup for the very first analysis
        self.out_graph = networkx.DiGraph(self._graph)
        if self._max_opt_iters > 1:
            self._fixed_point_analyze(cache=cache)
        else:
            updates = self._analyze(cache=cache)
            if not updates:
                self.out_graph = None

        # analysis is complete, no out_graph means it failed somewhere along the way
        if self.out_graph is None:
            return

        if not self._graph_is_structurable(self.out_graph):
            self.out_graph = None
            return

        # simplify the AIL graph
        if self._simplify_ail:
            # this should not (TM) change the structure of the graph but is needed for later optimizations
            self.out_graph = self._simplify_ail_graph(self.out_graph)

        if self._prevent_new_gotos:
            prev_gotos = len(initial_gotos)
            new_gotos = len(self._goto_manager.gotos)
            if (self._strictly_less_gotos and (new_gotos >= prev_gotos)) or (
                not self._strictly_less_gotos and (new_gotos > prev_gotos)
            ):
                self.out_graph = None
                return

    def _fixed_point_analyze(self, cache=None):
        for _ in range(self._max_opt_iters):
            # backup the graph before the optimization
            if self._recover_structure_fails and self.out_graph is not None:
                self._prev_graph = networkx.DiGraph(self.out_graph)

            # run the optimization, output applied to self.out_graph
            changes = self._analyze(cache=cache)
            if not changes:
                break

            # check if the graph is structurable
            if not self._graph_is_structurable(self.out_graph):
                self.out_graph = self._prev_graph if self._recover_structure_fails else None
                break

    def _simplify_ail_graph(self, graph):
        simp = self.project.analyses.AILSimplifier(
            self._func,
            func_graph=graph,
            gp=self._func.info.get("gp", None) if self.project.arch.name in {"MIPS32", "MIPS64"} else None,
        )
        return simp.func_graph if simp.simplified else graph

    def _graph_is_structurable(self, graph, readd_labels=False) -> bool:
        """
        Checks weather the input graph is structurable under the Phoenix schema-matching structuring algorithm.
        As a side effect, this will also update the region identifier and goto manager of this optimization pass.
        Consequently, a true return guarantees up-to-date goto information in the goto manager.
        """
        if readd_labels:
            graph = add_labels(graph)

        self._ri = self.project.analyses[RegionIdentifier].prep(kb=self.kb)(
            self._func,
            graph=graph,
            # never update the graph in-place, we need to keep the original graph for later use
            update_graph=False,
            cond_proc=self._ri.cond_proc,
            force_loop_single_exit=False,
            complete_successors=True,
        )
        if self._ri is None:
            return False

        rs = self.project.analyses[RecursiveStructurer].prep(kb=self.kb)(
            self._ri.region,
            cond_proc=self._ri.cond_proc,
            func=self._func,
            structurer_cls=PhoenixStructurer,
        )
        if not rs or not rs.result or not rs.result.nodes:
            return False

        rs = self.project.analyses.RegionSimplifier(self._func, rs.result, kb=self.kb, variable_kb=self._variable_kb)
        if not rs or rs.goto_manager is None:
            return False

        self._goto_manager = rs.goto_manager
        return True



import angr


class puts(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, string):
        stdout = self.state.posix.get_fd(1)
        if stdout is None:
            return -1

        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]
        length = self.inline_call(strlen, string).ret_expr
        out = stdout.write(string, length)
        stdout.write_data(self.state.solver.BVV(b"\n"))
        return (out + 1)[31:0]



import angr


class strcat(angr.SimProcedure):
    # pylint: disable=arguments-differ
    def run(self, dst, src):
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]
        strncpy = angr.SIM_PROCEDURES["libc"]["strncpy"]
        src_len = self.inline_call(strlen, src).ret_expr
        dst_len = self.inline_call(strlen, dst).ret_expr

        self.inline_call(strncpy, dst + dst_len, src, src_len + 1, src_len=src_len)
        return dst



from ..errors import AngrExitError
from . import ExplorationTechnique

import logging

l = logging.getLogger(name=__name__)


class Slicecutor(ExplorationTechnique):
    """
    The Slicecutor is an exploration that executes provided code slices.
    """

    def __init__(self, annotated_cfg, force_taking_exit=False, force_sat: bool = False):
        """
        All parameters except `annotated_cfg` are optional.

        :param annotated_cfg:       The AnnotatedCFG that provides the code slice.
        :param force_taking_exit:   Set to True if you want to create a successor based on our slice in case of
                                    unconstrained successors.
        :param force_sat:           If a branch specified by the slice is unsatisfiable, set this option to True if you
                                    want to force it to be satisfiable and be taken anyway.
        """
        super().__init__()

        self._annotated_cfg = annotated_cfg
        self._force_taking_exit = force_taking_exit
        self._force_sat = force_sat

    def setup(self, simgr):
        for stash in ("cut", "mysteries"):
            simgr.populate(stash, [])

    def filter(self, simgr, state, **kwargs):
        l.debug("Checking state %s for filtering...", state)
        return simgr.filter(state, **kwargs)

    def step_state(self, simgr, state, **kwargs):
        l.debug("%s ticking state %s at address %#x.", self, state, state.addr)
        stashes = simgr.step_state(state, **kwargs)

        new_active = []
        new_cut = []
        new_mystery = []

        # SimulationManager returns new active states in the None stash by default.
        flat_successors = stashes.get(None, None)
        if flat_successors is None:
            # Did the user explicitly put them into the 'active' stash instead?
            flat_successors = stashes.pop("active", [])

        for successor in flat_successors:
            l.debug("... checking exit to %#x from %#x.", successor.addr, state.addr)

            try:
                taken = self._annotated_cfg.should_take_exit(state.addr, successor.addr)
            except AngrExitError:  # TODO: which exception?
                l.debug("... annotated CFG did not know about it!")
                new_mystery.append(successor)
            else:
                if taken:
                    l.debug("... taking the exit.")
                    new_active.append(successor)
                    # the else case isn't here, because the state should set errored in this
                    # case and we'll catch it below
                else:
                    l.debug("... not taking the exit.")
                    new_cut.append(successor)

        unconstrained_successors = stashes.get("unconstrained", [])
        if not new_active and unconstrained_successors and self._force_taking_exit:
            stashes["unconstrained"] = []
            # somehow there is no feasible state. We are forced to create a successor based on our slice
            if len(unconstrained_successors) != 1:
                raise Exception("This should absolutely never happen, what?")
            for target in self._annotated_cfg.get_targets(state.addr):
                successor = unconstrained_successors[0].copy()
                successor.regs._ip = target
                new_active.append(successor)
            l.debug("Got unconstrained: %d new states are created based on AnnotatedCFG.", len(new_active))

        unsat_successors = stashes.get("unsat", None)
        if not new_active and unsat_successors and self._force_sat:
            stashes["unsat"] = []
            # find the state
            targets = self._annotated_cfg.get_targets(state.addr)
            if targets is None:
                targets = []
            for target in targets:
                try:
                    suc = next(iter(u for u in unsat_successors if u.addr == target))
                except StopIteration:
                    continue

                # drop all constraints
                if suc.mode == "fastpath":
                    # dropping constraints and making the state satisfiable again under fastpath mode is easy
                    suc.solver._solver.constraints = []
                    suc._satisfiable = True
                    new_active.append(suc)
                    l.debug("Forced unsat at %#x to be sat again.", suc.addr)
                else:
                    # with multiple possible solver frontends, dropping states in other state modes is not
                    # straightforward. I'll leave it to the next person who uses this feature
                    l.warning("force_sat is not implemented for solver mode %s.", suc.moe)

        stashes[None] = new_active
        stashes["mystery"] = new_mystery
        stashes["cut"] = new_cut
        return stashes

    def successors(self, simgr, state, **kwargs):
        kwargs["whitelist"] = self._annotated_cfg.get_whitelisted_statements(state.addr)
        kwargs["last_stmt"] = self._annotated_cfg.get_last_statement_index(state.addr)
        return simgr.successors(state, **kwargs)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("mi.dll")
prototypes = \
    {
        #
        'MI_Application_InitializeV1': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"ft": SimTypePointer(SimStruct({"Clone": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Destruct": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Delete": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "IsA": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "GetClassNameA": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetNameSpace": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "GetNameSpace": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "GetElementCount": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "AddElement": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetElement": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetElementAt": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "GetElement": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "GetElementAt": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "ClearElement": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "ClearElementAt": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "GetServerName": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "SetServerName": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "GetClass": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="MI_InstanceFT", pack=False, align=None), offset=0), "classDecl": SimTypePointer(SimStruct({"flags": SimTypeInt(signed=False, label="UInt32"), "code": SimTypeInt(signed=False, label="UInt32"), "name": SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), "qualifiers": SimTypePointer(SimTypePointer(SimTypeBottom(label="MI_Qualifier"), offset=0), offset=0), "numQualifiers": SimTypeInt(signed=False, label="UInt32"), "properties": SimTypePointer(SimTypePointer(SimTypeBottom(label="MI_PropertyDecl"), offset=0), offset=0), "numProperties": SimTypeInt(signed=False, label="UInt32"), "size": SimTypeInt(signed=False, label="UInt32"), "superClass": SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), "superClassDecl": SimTypePointer(SimTypeBottom(label="MI_ClassDecl"), offset=0), "methods": SimTypePointer(SimTypePointer(SimTypeBottom(label="MI_MethodDecl"), offset=0), offset=0), "numMethods": SimTypeInt(signed=False, label="UInt32"), "schema": SimTypePointer(SimTypeBottom(label="MI_SchemaDecl"), offset=0), "providerFT": SimTypePointer(SimTypeBottom(label="MI_ProviderFT"), offset=0), "owningClass": SimTypePointer(SimTypeBottom(label="MI_Class"), offset=0)}, name="MI_ClassDecl", pack=False, align=None), offset=0), "serverName": SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), "nameSpace": SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), "reserved": SimTypeFixedSizeArray(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), 4)}, name="MI_Instance", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimStruct({"reserved1": SimTypeLongLong(signed=False, label="UInt64"), "reserved2": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "ft": SimTypePointer(SimStruct({"Close": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewSession": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewHostedProvider": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewInstance": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewDestinationOptions": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewOperationOptions": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewSubscriptionDeliveryOptions": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewSerializer": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewDeserializer": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewInstanceFromClass": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "NewClass": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="MI_ApplicationFT", pack=False, align=None), offset=0)}, name="MI_Application", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="MI_Result"), arg_names=["flags", "applicationID", "extendedError", "application"]),
    }

lib.set_prototypes(prototypes)



import angr
from cle.backends.externs.simdata.io_file import io_file_data_for_arch

import logging

l = logging.getLogger(name=__name__)


class __getdelim(angr.SimProcedure):
    # this code is modified from the 'fgets' implementation
    #   to take an arbitrary delimiter
    #   with no max size for concrete data

    # pylint: disable=arguments-differ
    def run(self, line_ptrptr, len_ptr, delim, file_ptr):
        # let's get the memory back for the file we're interested in and find the delimiter
        fd_offset = io_file_data_for_arch(self.state.arch)["fd"]
        fd = self.state.mem[file_ptr + fd_offset :].int.resolved
        simfd = self.state.posix.get_fd(fd)
        if simfd is None:
            return -1

        # symbolic delimiters will make this tricky
        if delim.symbolic:
            raise angr.SimProcedureError("I don't know how to handle a symbolic delimiter")

        # case 1: the data is concrete. we should read it a byte at a time since we can't seek for
        # the newline and we don't have any notion of buffering in-memory
        if simfd.read_storage.concrete:
            if self.state.solver.is_true(simfd.eof()):
                # End-of-file reached
                return -1

            realloc = angr.SIM_PROCEDURES["libc"]["realloc"]

            # #dereference the destination buffer
            line_ptr = self.state.memory.load(line_ptrptr, 8)
            size = 120
            # im just always going to realloc and restart at size = 120, regardless of if a proper size buffer exists.
            # this doesn't match the exact behavior of get delim, but is the easiest way to ignore symbolic sizes.
            dst = self.inline_call(realloc, line_ptr, size).ret_expr

            count = 0
            while True:
                data, real_size = simfd.read_data(1)
                if self.state.solver.is_true(real_size == 0):
                    break
                self.state.memory.store(dst + count, data)
                count += 1
                if count == size:
                    size = count + size + 1
                    dst = self.inline_call(realloc, dst, size).ret_expr
                if delim.size() > data.size():
                    data = data.zero_extend(delim.size() - data.size())
                if self.state.solver.is_true(data == delim):
                    break

            self.state.memory.store(dst + count, b"\0")
            self.state.memory.store(len_ptr, count)
            self.state.memory.store(line_ptrptr, dst)
            return count

        # case 2: the data is symbolic, the delimiter could be anywhere. Read some maximum number of bytes
        # and add a constraint to assert the delimiter nonsense.
        # caveat: there could also be no delimiter and the file could EOF.
        else:
            # Just a guess as to a good value for a max size
            size = 1024

            data, real_size = simfd.read_data(size - 1)
            delim_byte = chr(self.state.solver.eval(delim))

            for i, byte in enumerate(data.chop(8)):
                self.state.add_constraints(
                    self.state.solver.If(
                        i + 1 != real_size,
                        byte != delim_byte,  # if not last byte returned, not newline
                        self.state.solver.Or(  # otherwise one of the following must be true:
                            i + 2 == size,  # - we ran out of space, or
                            simfd.eof(),  # - the file is at EOF, or
                            byte == delim_byte,  # - it is a newline
                        ),
                    )
                )

            malloc = angr.SIM_PROCEDURES["libc"]["malloc"]

            dst = self.inline_call(malloc, real_size).ret_expr

            self.state.memory.store(dst, data, size=real_size)
            self.state.memory.store(dst + real_size, b"\0")

            self.state.memory.store(len_ptr, real_size)
            self.state.memory.store(line_ptrptr, dst)

            return real_size



from .mips_elf_fast import MipsElfFastResolver
from .x86_elf_pic_plt import X86ElfPicPltResolver
from .jumptable import JumpTableResolver
from .x86_pe_iat import X86PeIatResolver
from .amd64_elf_got import AMD64ElfGotResolver
from .arm_elf_fast import ArmElfFastResolver
from .const_resolver import ConstantResolver
from .amd64_pe_iat import AMD64PeIatResolver



import angr


class xstrtol_fatal(angr.SimProcedure):
    """
    xstrtol_fatal
    """

    NO_RET = True

    # pylint: disable=unused-argument,arguments-differ
    def run(self, err, opt_idx, c, long_options, arg):
        self.exit(1)



from ailment.expression import Convert, BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ConvASub0ShrAnd(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "Conv(M->1, (expr >> N) & 1) => expr < 0"
    expr_classes = (Convert,)  # all expressions are allowed

    def optimize(self, expr: Convert, **kwargs):
        # Conv(M->1, ((expr) >> N) & 1) => expr < 0
        # Conv(M->1, ((expr - 0) >> N) & 1) => expr < 0
        if expr.to_bits == 1:
            if (
                isinstance(expr.operand, BinaryOp)
                and expr.operand.op == "And"
                and isinstance(expr.operand.operands[1], Const)
                and expr.operand.operands[1].value == 1
            ):
                # taking a single bit
                inner_expr = expr.operand.operands[0]
                if (
                    isinstance(inner_expr, BinaryOp)
                    and inner_expr.op == "Shr"
                    and isinstance(inner_expr.operands[1], Const)
                ):
                    # right-shifting with a constant
                    shr_amount = inner_expr.operands[1].value
                    if shr_amount == 7:
                        # int8_t
                        to_bits = 8
                    elif shr_amount == 15:
                        # int16_t
                        to_bits = 16
                    elif shr_amount == 31:
                        # int32_t
                        to_bits = 32
                    elif shr_amount == 63:
                        # int64_t
                        to_bits = 64
                    else:
                        # unsupported
                        return None

                    real_expr = inner_expr.operands[0]

                    if (
                        isinstance(real_expr, BinaryOp)
                        and real_expr.op == "Sub"
                        and isinstance(real_expr.operands[1], Const)
                        and real_expr.operands[1].value == 0
                    ):
                        real_expr = real_expr.operands[0]

                    cvt = Convert(expr.idx, real_expr.bits, to_bits, False, real_expr, **expr.tags)
                    cmp = BinaryOp(
                        None,
                        "CmpLT",
                        (
                            cvt,
                            Const(None, None, 0, to_bits),
                        ),
                        True,
                        **expr.tags,
                    )
                    return cmp

        return None



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("loadperf.dll")
prototypes = \
    {
        #
        'InstallPerfDllW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szComputerName", "lpIniFile", "dwFlags"]),
        #
        'InstallPerfDllA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szComputerName", "lpIniFile", "dwFlags"]),
        #
        'LoadPerfCounterTextStringsA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpCommandLine", "bQuietModeArg"]),
        #
        'LoadPerfCounterTextStringsW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpCommandLine", "bQuietModeArg"]),
        #
        'UnloadPerfCounterTextStringsW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpCommandLine", "bQuietModeArg"]),
        #
        'UnloadPerfCounterTextStringsA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpCommandLine", "bQuietModeArg"]),
        #
        'UpdatePerfNameFilesA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szNewCtrFilePath", "szNewHlpFilePath", "szLanguageID", "dwFlags"]),
        #
        'UpdatePerfNameFilesW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szNewCtrFilePath", "szNewHlpFilePath", "szLanguageID", "dwFlags"]),
        #
        'SetServiceAsTrustedA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szReserved", "szServiceName"]),
        #
        'SetServiceAsTrustedW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szReserved", "szServiceName"]),
        #
        'BackupPerfRegistryToFileW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szFileName", "szCommentString"]),
        #
        'RestorePerfRegistryFromFileW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szFileName", "szLangId"]),
    }

lib.set_prototypes(prototypes)



import claripy

from .structuring.structurer_nodes import ConditionNode, CodeNode
from .sequence_walker import SequenceWalker


class JumpTableEntryConditionRewriter(SequenceWalker):
    """
    Remove artificial jump table entry conditions that ConditionProcessor introduced when dealing with jump tables.
    """

    def __init__(self, jumptable_entry_conds):
        super().__init__()
        self._jumptable_entry_conds = jumptable_entry_conds

    def _process_expr(self, expr):
        if expr in self._jumptable_entry_conds:
            return claripy.true

        new_args = []
        replaced = False
        if expr.op in {"Or", "And", "Not"}:
            for arg in expr.args:
                new_arg = self._process_expr(arg)
                if new_arg is not None:
                    replaced = True
                new_args.append(new_arg if new_arg is not None else arg)
        if replaced:
            return getattr(claripy, expr.op)(*new_args)
        return None

    def _handle_Code(self, node: CodeNode, **kwargs):
        new_node = super()._handle_Code(node, **kwargs)
        changed = False
        if new_node is not None:
            changed = True
        else:
            new_node = node

        new_cond = self._process_expr(new_node.reaching_condition) if new_node.reaching_condition is not None else None
        if new_cond is not None:
            changed = True
        else:
            new_cond = new_node.reaching_condition

        if changed:
            return CodeNode(new_node.node, new_cond)
        return None

    def _handle_Condition(self, node: ConditionNode, **kwargs):
        new_node = super()._handle_Condition(node, **kwargs)
        changed = False
        if new_node is None:
            new_node = node
        else:
            changed = True

        new_cond = self._process_expr(new_node.condition)
        if new_cond is not None:
            return ConditionNode(
                new_node.addr, node.reaching_condition, new_cond, new_node.true_node, false_node=new_node.false_node
            )

        if changed:
            return new_node
        return None



# pylint:disable=unused-import
from ..models import DbCFGModel
from ...knowledge_plugins.cfg.cfg_model import CFGModel


class CFGModelSerializer:
    """
    Serialize/unserialize a CFGModel.
    """

    @staticmethod
    def dump(session, db_kb, ident, cfg_model):
        """

        :param session:
        :param DbKnowledgeBase db_kb:   The database object for KnowledgeBase.
        :param str ident:               Identifier of the CFG model.
        :param CFGModel cfg_model:      The CFG model to dump.
        :return:                        None
        """

        db_cfg_id = session.query(DbCFGModel.id).filter_by(ident=ident).scalar()
        if db_cfg_id is not None:
            # remove the existing CFG
            session.query(DbCFGModel).filter_by(id=db_cfg_id).delete()

        db_cfg = DbCFGModel(
            kb=db_kb,
            ident=ident,
            blob=cfg_model.serialize(),
        )
        session.add(db_cfg)

    @staticmethod
    def load(session, db_kb, ident, cfg_manager, loader=None):
        db_cfg: DbCFGModel = session.query(DbCFGModel).filter_by(kb=db_kb, ident=ident).scalar()
        if db_cfg is None:
            return None

        cfg_model = CFGModel.parse(db_cfg.blob, cfg_manager=cfg_manager, loader=loader)
        return cfg_model



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("api-ms-win-net-isolation-l1-1-0.dll")
prototypes = \
    {
        #
        'NetworkIsolationSetupAppContainerBinaries': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["applicationContainerSid", "packageFullName", "packageFolder", "displayName", "bBinariesFullyComputed", "binaries", "binariesCount"]),
        #
        'NetworkIsolationRegisterForAppContainerChanges': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"changeType": SimTypeInt(signed=False, label="INET_FIREWALL_AC_CHANGE_TYPE"), "createType": SimTypeInt(signed=False, label="INET_FIREWALL_AC_CREATION_TYPE"), "appContainerSid": SimTypePointer(SimTypeBottom(label="SID"), offset=0), "userSid": SimTypePointer(SimTypeBottom(label="SID"), offset=0), "displayName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Anonymous": SimUnion({"capabilities": SimStruct({"count": SimTypeInt(signed=False, label="UInt32"), "capabilities": SimTypePointer(SimTypeBottom(label="SID_AND_ATTRIBUTES"), offset=0)}, name="INET_FIREWALL_AC_CAPABILITIES", pack=False, align=None), "binaries": SimStruct({"count": SimTypeInt(signed=False, label="UInt32"), "binaries": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)}, name="INET_FIREWALL_AC_BINARIES", pack=False, align=None)}, name="<anon>", label="None")}, name="INET_FIREWALL_AC_CHANGE", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["context", "pChange"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["flags", "callback", "context", "registrationObject"]),
        #
        'NetworkIsolationUnregisterForAppContainerChanges': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["registrationObject"]),
        #
        'NetworkIsolationFreeAppContainers': SimTypeFunction([SimTypePointer(SimStruct({"appContainerSid": SimTypePointer(SimTypeBottom(label="SID"), offset=0), "userSid": SimTypePointer(SimTypeBottom(label="SID"), offset=0), "appContainerName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "displayName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "description": SimTypePointer(SimTypeChar(label="Char"), offset=0), "capabilities": SimStruct({"count": SimTypeInt(signed=False, label="UInt32"), "capabilities": SimTypePointer(SimTypeBottom(label="SID_AND_ATTRIBUTES"), offset=0)}, name="INET_FIREWALL_AC_CAPABILITIES", pack=False, align=None), "binaries": SimStruct({"count": SimTypeInt(signed=False, label="UInt32"), "binaries": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)}, name="INET_FIREWALL_AC_BINARIES", pack=False, align=None), "workingDirectory": SimTypePointer(SimTypeChar(label="Char"), offset=0), "packageFullName": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="INET_FIREWALL_APP_CONTAINER", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pPublicAppCs"]),
        #
        'NetworkIsolationEnumAppContainers': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"appContainerSid": SimTypePointer(SimTypeBottom(label="SID"), offset=0), "userSid": SimTypePointer(SimTypeBottom(label="SID"), offset=0), "appContainerName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "displayName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "description": SimTypePointer(SimTypeChar(label="Char"), offset=0), "capabilities": SimStruct({"count": SimTypeInt(signed=False, label="UInt32"), "capabilities": SimTypePointer(SimTypeBottom(label="SID_AND_ATTRIBUTES"), offset=0)}, name="INET_FIREWALL_AC_CAPABILITIES", pack=False, align=None), "binaries": SimStruct({"count": SimTypeInt(signed=False, label="UInt32"), "binaries": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)}, name="INET_FIREWALL_AC_BINARIES", pack=False, align=None), "workingDirectory": SimTypePointer(SimTypeChar(label="Char"), offset=0), "packageFullName": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="INET_FIREWALL_APP_CONTAINER", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "pdwNumPublicAppCs", "ppPublicAppCs"]),
        #
        'NetworkIsolationGetAppContainerConfig': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"Sid": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Attributes": SimTypeInt(signed=False, label="UInt32")}, name="SID_AND_ATTRIBUTES", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pdwNumPublicAppCs", "appContainerSids"]),
        #
        'NetworkIsolationSetAppContainerConfig': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Sid": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Attributes": SimTypeInt(signed=False, label="UInt32")}, name="SID_AND_ATTRIBUTES", pack=False, align=None), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwNumPublicAppCs", "appContainerSids"]),
        #
        'NetworkIsolationDiagnoseConnectFailureAndGetInfo': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="NETISO_ERROR_TYPE"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["wszServerName", "netIsoError"]),
    }

lib.set_prototypes(prototypes)



from angr.utils.graph import GraphUtils
from .graph import GraphVisitor


class LoopVisitor(GraphVisitor):
    """
    :param angr.analyses.loopfinder.Loop loop: The loop to visit.
    """

    def __init__(self, loop):
        super().__init__()
        self.loop = loop

        self.reset()

    def successors(self, node):
        return self.loop.graph.successors(node)

    def predecessors(self, node):
        return self.loop.graph.predecessors(node)

    def sort_nodes(self, nodes=None):
        sorted_nodes = GraphUtils.quasi_topological_sort_nodes(self.loop.graph)

        if nodes is not None:
            sorted_nodes = [n for n in sorted_nodes if n in set(nodes)]

        return sorted_nodes



import angr


class uname(angr.SimProcedure):
    def run(self, uname_buf):  # pylint: disable=arguments-differ
        # struct utsname {
        #     char sysname[];    /* Operating system name (e.g., "Linux") */
        #     char nodename[];   /* Name within "some implementation-defined
        #                           network" */
        #     char release[];    /* Operating system release (e.g., "2.6.28") */
        #     char version[];    /* Operating system version */
        #     char machine[];    /* Hardware identifier */
        # };

        off = self._store(uname_buf, b"Linux", 0)
        off += self._store(uname_buf, b"localhost", off)
        off += self._store(uname_buf, b"4.0.0", off)
        off += self._store(uname_buf, b"#1 SMP Mon Jan 01 00:00:00 GMT 1970", off)
        if self.state.arch.bits == 64:
            self._store(uname_buf, b"x86_64", off)
        else:
            self._store(uname_buf, b"x86", off)

        return 0  # success

    def _store(self, uname_buf, val, off):
        self.state.memory.store(uname_buf + off, val + (b"\0" * (65 - len(val))))
        return 65
